import sys, md5, time, threading, socket, Queue, signal, sqlite, os, re
import textwrap, urllib2, urlparse, HTMLParser
import param, feedparser, normalize, util, transform

socket.setdefaulttimeout(10)
feedparser.USER_AGENT = param.user_agent

def escape(str):
  return str.replace("'", "''")

class ParseError(Exception):
  pass
class AutodiscoveryParseError(Exception):
  pass
class FeedAlreadyExists(Exception):
  pass
class UnknownError(Exception):
  pass

class AutoDiscoveryHandler(HTMLParser.HTMLParser):
  """Find RSS autodiscovery info, as specified in:
    http://diveintomark.org/archives/2002/05/30/rss_autodiscovery
  Cope even if the HTML document is not strictly XML compliant (as we do not
  use a SGML parser like htmllib.HTMLParser does"""
  def __init__(self):
    HTMLParser.HTMLParser.__init__(self)
    self.autodiscovery = {}
  def handle_starttag(self, tag, attrs):
    if tag == 'link':
      attrs = dict(attrs)
      if attrs.get('rel', '').strip().lower() != 'alternate':
        return
      if attrs.get('type') == 'application/rss+xml' and 'href' in attrs:
        self.autodiscovery['rss'] = attrs['href']
      if attrs.get('type') == 'application/atom+xml' and 'href' in attrs:
        self.autodiscovery['atom'] = attrs['href']
  def feed_url(self, page_url):
    page_data = urllib2.urlopen(page_url).read()
    self.feed(page_data)
    # Atom has cleaner semantics than RSS, so give it priority
    url = self.autodiscovery.get(
      'atom', self.autodiscovery.get('rss'))
    # the URL could be relative, if so fix it
    url = urlparse.urljoin(page_url, url)
    return url

def re_autodiscovery(url):
  autodiscovery_re = re.compile(
    '<link[^>]*rel="alternate"[^>]*'
    'application/(atom|rss)\\+xml[^>]*href="([^"]*)"')
  candidates = autodiscovery_re.findall(urllib2.urlopen(url).read())
  candidates.sort()
  return candidates

def add_feed(feed_xml):
  """Try to add a feed. Returns a tuple (feed_uid, num_added, num_filtered)"""
  from singleton import db
  c = db.cursor()
  try:
    # verify the feed
    f = feedparser.parse(feed_xml)
    # CVS versions of feedparser are not throwing exceptions as they should
    # see:
    # http://sourceforge.net/tracker/index.php?func=detail&aid=1379172&group_id=112328&atid=661937
    if not f.feed or ('link' not in f.feed or 'title' not in f.feed):
      # try autodiscovery
      try:
        feed_xml = AutoDiscoveryHandler().feed_url(feed_xml)
      except HTMLParser.HTMLParseError:
        # in desperate conditions, regexps ride to the rescue
        try:
          feed_xml = re_autodiscovery(feed_xml)[0][1]
        except:
          raise AutodiscoveryParseError
      if not feed_xml:
        raise ParseError
      f = feedparser.parse(feed_xml)
      if not f.feed:
        raise ParseError
    # we have a valid feed, normalize it
    normalize.normalize_feed(f)
    feed = {
      'xmlUrl': f['url'],
      'htmlUrl': str(f.feed['link']),
      'etag': f['etag'],
      'title': f.feed['title'].encode('ascii', 'xmlcharrefreplace'),
      'desc': f.feed['description'].encode('ascii', 'xmlcharrefreplace')
      }
    for key, value in feed.items():
      if type(value) == str:
        feed[key] = escape(value)
    load_rules()
    try:
      c.execute("""insert into fm_feeds
      (feed_xml, feed_etag, feed_html, feed_title, feed_desc) values
      ('%(xmlUrl)s', '%(etag)s', '%(htmlUrl)s', '%(title)s',
      '%(desc)s')""" % feed)
      feed_uid = c.lastrowid
      num_added, num_filtered = process_parsed_feed(f, c, feed_uid)
      db.commit()
      return (feed_uid, num_added, num_filtered)
    except sqlite.IntegrityError, e:
      if 'feed_xml' in str(e):
        db.rollback()
        raise FeedAlreadyExists
      else:
        db.rollback()
        raise UnknownError(str(e))
  finally:
    c.close()

def update_feed_xml(feed_uid, feed_xml):
  """Update a feed URL and fetch the feed. Returns the number of new items"""
  feed_uid = int(feed_uid)

  f = feedparser.parse(feed_xml)
  if not f.feed:
    raise ParseError
  normalize.normalize_feed(f)

  from singleton import db
  c = db.cursor()
  clear_errors(db, c, feed_uid, f)
  try:
    try:
      c.execute("""update fm_feeds set feed_xml='%s', feed_html='%s'
      where feed_uid=%d""" \
                % (escape(feed_xml), escape(str(f.feed['link'])), feed_uid))
    except sqlite.IntegrityError, e:
      if 'feed_xml' in str(e):
        db.rollback()
        raise FeedAlreadyExists
      else:
        db.rollback()
        raise UnknownError(str(e))
    load_rules()
    num_added = process_parsed_feed(f, c, feed_uid)
    db.commit()
    return num_added
  finally:
    c.close()

def update_feed_title(feed_uid, feed_title):
  """Update a feed title"""
  feed_uid = int(feed_uid)

  from singleton import db
  c = db.cursor()
  try:
    c.execute("""update fm_feeds set feed_title='%s' where feed_uid=%d""" \
              % (escape(feed_title), feed_uid))
    db.commit()
  finally:
    c.close()

def update_feed_desc(feed_uid, feed_desc):
  """Update a feed desc"""
  feed_uid = int(feed_uid)

  from singleton import db
  c = db.cursor()
  try:
    c.execute("""update fm_feeds set feed_desc='%s' where feed_uid=%d""" \
              % (escape(feed_desc), feed_uid))
    db.commit()
  finally:
    c.close()

def update_feed_private(feed_uid, private):
  feed_uid = int(feed_uid)
  private = int(bool(private))
  from singleton import db
  c = db.cursor()
  try:
    c.execute("""update fm_feeds set feed_private=%d where feed_uid=%d""" \
              % (private, feed_uid))
    db.commit()
  finally:
    c.close()

def update_feed_dupcheck(feed_uid, dupcheck):
  feed_uid = int(feed_uid)
  dupcheck = int(bool(dupcheck))
  # XXX run a dupcheck pass retroactively here if dupcheck == 1
  from singleton import db
  c = db.cursor()
  try:
    c.execute("""update fm_feeds set feed_dupcheck=%d where feed_uid=%d""" \
              % (dupcheck, feed_uid))
    db.commit()
  finally:
    c.close()

def catch_up(feed_uid):
  feed_uid = int(feed_uid)
  from singleton import db
  c = db.cursor()
  try:
    c.execute("""update fm_items set item_rating=-1
    where item_feed_uid=%d and item_rating=0""" % feed_uid)
    db.commit()
  finally:
    c.close()

def purge_reload(feed_uid):
  feed_uid = int(feed_uid)
  from singleton import db
  c = db.cursor()
  try:
    c.execute("""delete from fm_items
    where item_feed_uid=%d and item_rating=0""" % feed_uid)
    c.execute("""update fm_feeds set feed_modified=NULL, feed_etag=NULL
    where feed_uid=%d""" % feed_uid)
    c.execute("""select feed_xml from fm_feeds
    where feed_uid=%d""" % feed_uid)
    feed_xml = c.fetchone()[0]
    db.commit()
    f = feedparser.parse(feed_xml)
    if not f.feed:
      raise ParseError
    normalize.normalize_feed(f)
    clear_errors(db, c, feed_uid, f)
    load_rules()
    num_added = process_parsed_feed(f, c, feed_uid)
    db.commit()
  finally:
    c.close()

def hard_purge(feed_uid):
  feed_uid = int(feed_uid)
  from singleton import db
  c = db.cursor()
  try:
    c.execute("""delete from fm_items
    where item_feed_uid=%d""" % feed_uid)
    c.execute("""delete from fm_feeds
    where feed_uid=%d""" % feed_uid)
    db.commit()
  finally:
    c.close()

def set_status(feed_uid, status):
  feed_uid = int(feed_uid)
  status = int(status)
  from singleton import db
  c = db.cursor()
  try:
    c.execute("""update fm_feeds set feed_status=%d where feed_uid=%d""" \
              % (status, feed_uid))
    db.commit()
  finally:
    c.close()

class FeedWorker(threading.Thread):
  def __init__(self, id, in_q, out_q):
    threading.Thread.__init__(self)
    self.id = id
    self.in_q = in_q
    self.out_q = out_q
    # we need to do this so temboz --refresh honors Ctrl-C
    self.setDaemon(True)
  def run(self):
    try:
      while True:
        feed = self.in_q.get()
        if not feed: return
        self.out_q.put((self.fetch_feed(*feed),) + feed)
    finally:
      self.out_q.put(None)
  def fetch_feed(self, feed_uid, feed_xml, feed_etag, feed_modified,
                 feed_dupcheck):
    print >> param.log, self.id, feed_xml
    return fetch_feed(feed_uid, feed_xml, feed_etag, feed_modified)

def fetch_feed(feed_uid, feed_xml, feed_etag, feed_modified):
  if not feed_etag:
    feed_etag = None
  if not feed_modified:
    feed_modified = None
  try:
    f = feedparser.parse(feed_xml, etag=feed_etag, modified=feed_modified)
  except socket.timeout:
    if param.debug:
      print >> param.log, 'EEEEE error fetching feed', feed_xml
    f = {'channel': {}, 'items': []}
  return f

def increment_errors(db, c, feed_uid):
  """Increment the error counter, and suspend the feed if the threshold is
  reached"""
  c.execute('update fm_feeds set feed_errors=feed_errors+1 where feed_uid=%d' \
            % feed_uid)
  c.execute("""select feed_errors, feed_title
  from fm_feeds where feed_uid=%d""" % feed_uid)
  errors, feed_title = c.fetchone()
  max_errors = getattr(param, 'max_errors', 100)
  if max_errors != -1 and errors > max_errors:
    # XXX we should generate a service announcement item if this happens
    print >> param.log, 'EEEEE too many errors, suspending feed', feed_title
    c.execute('update fm_feeds set feed_status = 1 where feed_uid=%d' \
              % feed_uid)

def clear_errors(db, c, feed_uid, f):
  'On successful feed parse, reset etag and/or modified date and error count'
  stmt = 'update fm_feeds set feed_errors=0'
  if 'etag' in f and f['etag']:
    stmt += ", feed_etag='%s'" % escape(f['etag'])
  else:
    stmt += ", feed_etag=NULL"
  if 'modified' in f and f['modified']:
    stmt += ", feed_modified=julianday(%f, 'unixepoch')" \
            % time.mktime(f['modified'])
  else:
    stmt += ", feed_modified=NULL"
  stmt += " where feed_uid=%d" % feed_uid
  c.execute(stmt)

def update_feed(db, c, f, feed_uid, feed_xml, feed_etag, feed_modified,
                feed_dupcheck=None):
  print >> param.log, feed_xml
  # check for errors - HTTP code 304 means no change
  if 'title' not in f.feed and 'link' not in f.feed and \
         ('status' not in f or f['status'] not in [304]):
    # error or timeout - increment error count
    increment_errors(db, c, feed_uid)
  else:
    # no error - reset etag and/or modified date and error count
    clear_errors(db, c, feed_uid, f)
  try:
    process_parsed_feed(f, c, feed_uid, feed_dupcheck)
  except:
    util.print_stack(['c', 'f'])

# shades of LISP...
def curry(fn, obj):
  return lambda *args: fn(obj, *args)

# obj can be a string, list or dictionary
def any(obj, *words):
  for w in words:
    if w in obj:
      return True
  return False

def union_any(obj_list, *words):
  for w in words:
    for obj in obj_list:
      if w in obj:
        return True
  return False

def process_parsed_feed(f, c, feed_uid, feed_dupcheck=None):
  """Insert the entries from a feedparser parsed feed f in the database using
the cursor c for feed feed_uid.
Returns a tuple (number of items added unread, number of filtered items)"""
  num_added = 0
  num_filtered = 0
  # check if duplicate title checking is in effect
  if feed_dupcheck is None:
    c.execute('select feed_dupcheck from fm_feeds where feed_uid=%d' \
              % feed_uid)
    feed_dupcheck = bool(c.fetchone()[0])
  # the Radio convention is reverse chronological order
  f['items'].reverse()
  for item in f['items']:
    normalize.normalize(item, f)
    skip = 0
    filter_dict = {}
    for key in f.feed:
      try:
        filter_dict['feed_' + key] = f.feed[key]
      except KeyError:
        pass
    filter_dict.update(item)
    # convenient shortcut functions
    filter_dict['title_any_words'] = curry(any, item['title_words'])
    filter_dict['content_any_words'] = curry(any, item['content_words'])
    filter_dict['union_any_words'] = curry(
      union_any, [item['title_words'], item['content_words']])
    filter_dict['title_any'] = curry(any, item['title'])
    filter_dict['content_any'] = curry(any, item['content'])
    filter_dict['union_any'] = curry(
      union_any, [item['title'], item['content']])
    filter_dict['title_any_lc'] = curry(any, item['title_lc'])
    filter_dict['content_any_lc'] = curry(any, item['content_lc'])
    filter_dict['union_any_lc'] = curry(
      union_any, [item['title_lc'], item['content_lc']])
    # evaluate the rules
    for rule in rules:
      try:
        skip = eval(rule, filter_dict)
        if skip:
          break
      except:
        util.print_stack(['f'])
    if skip:
      skip = -2
    title   = item['title']
    link    = item['link']
    guid    = item['id']
    author = item['author']
    created = item['created']
    modified = item['modified']
    if modified:
      modified = "julianday('%s')" % escape(modified)
    else:
      modified = 'NULL'
    content = item['content']
    # check if the item already exists, using the GUID as key
    c.execute("""select item_uid, item_link,
               item_loaded, item_created, item_modified,
               item_viewed, item_md5hex, item_title, item_content, item_creator
               from fm_items
               where item_feed_uid=%d and item_guid='%s'""" \
               % (feed_uid, escape(guid)))
    l = c.fetchall()
    # unknown GUID, but title duplicate checking may be in effect
    if not l:
      if feed_dupcheck:
        c.execute("""select count(*) from fm_items
        where item_feed_uid=%d and item_title='%s'""" \
                  % (feed_uid, escape(title)))
        l = bool(c.fetchone()[0])
        if l:
          print >> param.log, 'DUPLICATE TITLE', title
    # GUID already exists, this is a change
    else:
      assert len(l) == 1
      (item_uid, item_link, item_loaded, item_created, item_modified,
       item_viewed, item_md5hex, item_title, item_content, item_creator) = l[0]
      # if this is a feed without timestamps, use our timestamp to determine
      # the oldest item in the feed XML file
      if 'oldest' in f and f['oldest'] == '1970-01-01 00:00:00':
        if 'oldest_ts' not in f:
          f['oldest_ts'] = item_created
        else:
          f['oldest_ts'] = min(f['oldest_ts'], item_created)
      # XXX update item here
    # GUID doesn't exist yet, insert it
    if not l:
      try:
        sql = """insert into fm_items (item_feed_uid, item_guid,
        item_created,   item_modified, item_viewed, item_link, item_md5hex,
        item_title, item_content, item_creator, item_rating) values (%d, '%s',
        julianday('%s'), %s,          NULL,        '%s',      '%s',
        '%s',       '%s',         '%s', %d)""" % \
        (feed_uid, escape(guid), escape(created), modified, escape(link),
         md5.new(content).hexdigest(),
         escape(title),
         escape(content),
         escape(author),
         skip)
        c.execute(sql)
        if skip:
          num_filtered += 1
          print >> param.log, 'SKIP', title
        else:
          num_added += 1
          print >> param.log, ' ' * 4, title
      except:
        util.print_stack(['c', 'f'])
        continue
  # update timestamp of the oldest item still in the feed file
  if 'oldest' in f and f['oldest'] != '9999-99-99 99:99:99':
    if f['oldest'] == '1970-01-01 00:00:00' and 'oldest_ts' in f:
      c.execute("""update fm_feeds
      set feed_oldest=%s
      where feed_uid=%d""" % (f['oldest_ts'], feed_uid))
    else:
      c.execute("""update fm_feeds
      set feed_oldest=julianday('%s')
      where feed_uid=%d""" % (f['oldest'], feed_uid))
  
  return (num_added, num_filtered)

def cleanup(db=None, c=None):
  """garbage collection - see param.py
  this is done only once a day between 3 and 4 AM as this is quite intensive
  and could interfere with user activity
  It can also be invoked by running temboz --clean
  """
  if not db:
    from singleton import db
    c = db.cursor()
  from singleton import sqlite_cli
  if getattr(param, 'garbage_contents', False):
    c.execute("""update fm_items
    set item_content=''
    where item_rating<0 and item_created < julianday('now')-%d""" %
               param.garbage_contents)
    db.commit()
  if getattr(param, 'garbage_items', False):
    c.execute("""delete from fm_items where item_uid in (
      select item_uid from fm_items, fm_feeds
      where item_created < min(julianday('now') - %d, feed_oldest - 7)
      and item_rating < 0 and feed_uid=item_feed_uid)""" % param.garbage_items)
    db.commit()
  c.execute('vacuum')
  # we still hold the PseudoCursor lock, this is a good opportunity to backup
  try:
    os.mkdir('backups')
  except OSError:
    pass
  os.system((sqlite_cli + ' rss.db .dump | %s > backups/daily_' \
             + time.strftime('%Y-%m-%d') + '%s') % param.backup_compressor)
  # delete old backups
  backup_re = re.compile(
    'daily_[0-9][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9]\\.')
  for fn in os.listdir('backups'):
    m = backup_re.match(fn)
    if m:
      elapsed = time.time() - os.stat('backups/' + fn).st_ctime
      if elapsed > 86400 * param.daily_backups:
        try:
          os.remove('backups/' + fn)
        except OSError:
          pass
  
def update():
  from singleton import db
  c = db.cursor()
  # refresh filtering rules
  load_rules()
  # at 3AM by default, perform house-cleaning
  if time.localtime()[3] == param.backup_hour:
    cleanup(db, c)
  # create worker threads and the queues used to communicate with them
  work_q = Queue.Queue()
  process_q = Queue.Queue()
  workers = []
  for i in range(param.feed_concurrency):
    workers.append(FeedWorker(i + 1, work_q, process_q))
    workers[-1].start()
  # assign work
  c.execute("""select feed_uid, feed_xml, feed_etag, feed_dupcheck,
  strftime('%s', feed_modified)
  from fm_feeds where feed_status=0""")
  for feed_uid, feed_xml, feed_etag, feed_dupcheck, feed_modified in c:
    if feed_modified:
      feed_modified = float(feed_modified)
      feed_modified = time.localtime(feed_modified)
    else:
      feed_modified = None
    work_q.put((feed_uid, feed_xml, feed_etag, feed_modified, feed_dupcheck))
  # None is an indication for workers to stop
  for i in range(param.feed_concurrency):
    work_q.put(None)
  workers_left = param.feed_concurrency
  while workers_left > 0:
    feed_info = process_q.get()
    # exited worker
    if not feed_info:
      workers_left -= 1
    else:
      try:
        update_feed(db, c, *feed_info)
      except:
        util.print_stack()
  db.commit()
  c.close()

class PeriodicUpdater(threading.Thread):
  def __init__(self):
    threading.Thread.__init__(self)
    self.setDaemon(True)
  def run(self):
    while True:
      # XXX should wrap this in a try/except clause
      time.sleep(param.refresh_interval)
      print >> param.log, time.ctime(), '- refreshing feeds'
      try:
        update()
      except:
        util.print_stack()

##############################################################################
#
rules = []

rule_comment_re = re.compile('^#.*$', re.MULTILINE)
def normalize_rule(rule):
  """allow embedded CR/LF and comments to make for more readable rules"""
  return rule_comment_re.sub('', rule).replace(
    '\n', ' ').replace('\r', ' ').strip()

wrapper = textwrap.TextWrapper(width=80, break_long_words=False)
# XXX this relies on texwrap implementation details to prevent wrapping on
# XXX hyphens and em-dashes, only on spaces
wrapper.wordsep_re = re.compile(r'(\s+)')
def rule_lines(rule):
  "Find how many lines are needed for the rule in a word-wrapped <textarea>"
  lines = 0
  for line in rule.splitlines():
    if line.strip():
      lines += len(wrapper.wrap(line))
    else:
      lines += 1
  return lines

def load_rules():
  global rules
  rules = []
  from singleton import db
  c = db.cursor()
  try:
    c.execute("""select rule_uid, rule_text
    from fm_rules where rule_expires is null
    or rule_expires > julianday('now')""")
    for uid, rule in c:
      rule = normalize_rule(rule)
      rules.append(compile(rule, 'rule' + `uid`, 'eval'))
  finally:
    c.close()
  # reload the degunking filters as well
  reload(transform)

def update_rule(db, c, uid, expires, text, delete):
  if expires == 'never':
    expires = 'NULL'
  else:
    expires = "julianday('%s')" % expires
  # check syntax
  compile(normalize_rule(text), 'web form', 'eval')
  if uid == 'new':
    c.execute("""insert into fm_rules (rule_expires, rule_text)
    values (%s, '%s')""" \
              % (expires, escape(text)))
  elif delete == 'on':
    c.execute("delete from fm_rules where rule_uid=%s" % uid)
  else:
    c.execute("""update fm_rules set rule_expires=%s, rule_text='%s'
    where rule_uid=%s""" % (expires, escape(text), uid))
  db.commit()
    
